/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.configuration;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration.ContainerState;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfigurationListener;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.JndiConfiguration;
import org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerAlreadyExistException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerDoesNotExistException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerNameDuplicatedException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerDuplicatedPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerFastRemoteTcpPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerHostException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerJmxPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerNameException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerPasswordException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerTransportSvcException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerUserException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentCurrentContainerMovedException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentCurrentContainerNoMoreExistException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentCurrentSubdomainNoMoreExistException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiFactoryException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiProviderUrlException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentNetworkTimeSyncException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentSubdomainConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentSubdomainModeException;
import org.ow2.petals.microkernel.api.configuration.exception.InvalidContainerConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.InvalidSubdomainConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.MissingServerPropertyException;
import org.ow2.petals.microkernel.api.configuration.exception.NoContainerConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoDomainConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoDomainNameProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoServerPropertiesProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoSubdomainConfigurationProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoSubdomainNameProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.NoTopologyProvidedException;
import org.ow2.petals.microkernel.api.configuration.exception.SubdomainAlreadyExistException;
import org.ow2.petals.microkernel.api.configuration.exception.SubdomainDoesNotExistException;
import org.ow2.petals.microkernel.api.configuration.exception.SubdomainNameDuplicatedException;
import org.ow2.petals.microkernel.api.configuration.exception.SubdomainUnknownException;
import org.ow2.petals.microkernel.api.jbi.messaging.RouterService;
import org.ow2.petals.microkernel.api.system.persistence.PersistenceService;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.topology.TopologyBuilder;
import org.ow2.petals.topology.TopologyException;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.Jndi;
import org.ow2.petals.topology.generated.Subdomain;
import org.ow2.petals.topology.generated.SubdomainMode;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;
import com.ebmwebsourcing.easycommons.properties.PropertiesHelper;

/**
 * <p>
 * The Configuration service that handles the configuration of the local PEtALS
 * container.
 * </p>
 * <p>
 * At its starting, it retrieves the configuration from the {@link Topology} and
 * the local container configuration (ie, content of the file
 * 'server.properties'), it performs a first level validation of the
 * consistency, then it fill the local {@link ContainerConfiguration}.
 * </p>
 * 
 * @author Roland NAUDIN - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Charles Casadei - EBM WebSourcing
 * @since Petals 2.0
 */
@Component(provides = @Interface(name = "service", signature = ConfigurationService.class))
public class ConfigurationServiceImpl implements ConfigurationService {

    /**
     * File name, in the working area, of the persisted topology definition
     */
    private static final String PERSISTED_TOPOLOGY_DEFINITION = "topology.xml";

    /**
     * Logging utilities
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(ConfigurationService.COMPONENT_LOGGER_NAME));

    /**
     * The configuration of the current container
     */
    private final ContainerConfiguration containerConfiguration = new ContainerConfiguration();

    /**
     * The domain configuration of the domain for which the current container is a member.
     */
    private final DomainConfiguration domainConfiguration = new DomainConfiguration();

    /**
     * The sub-domain configuration of the sub-domain for which the current container is a member.
     */
    private final SubDomainConfiguration subDomainConfiguration = new SubDomainConfiguration();

    /**
     * List of sub-domain configuration created from the topology. Key is the sub-domain name.
     */
    private final ConcurrentMap<String, SubDomainConfiguration> subDomainsConfig = new ConcurrentHashMap<String, SubDomainConfiguration>();

    /**
     * List of remote container configuration created from the topology. Key is the container name.
     */
    private final ConcurrentMap<String, ContainerConfiguration> containersConfig = new ConcurrentHashMap<String, ContainerConfiguration>();

    /**
     * The initial local topology read from the local topology definition (ie. topology.xml)
     */
    private Topology initialLocalTopology;

    private final List<ContainerConfigurationListener> containerConfigListeners = Collections
            .synchronizedList(new ArrayList<ContainerConfigurationListener>());

    /**
     * {@inheritDoc}
     */
    public void addContainerConfiguration(final ContainerConfiguration addedContainer)
            throws NoContainerConfigurationProvidedException, InvalidContainerConfigurationProvidedException,
            ContainerAlreadyExistException, SubdomainUnknownException {

        if (addedContainer == null) {
            throw new NoContainerConfigurationProvidedException();
        }

        final String localContainerName = this.containerConfiguration.getName();

        // Retrieve the container name
        final String addedContainerName = addedContainer.getName();
        if (addedContainerName == null) {
            throw new InvalidContainerConfigurationProvidedException("The container name is not defined");
        }

        if (addedContainerName.equals(localContainerName)) {
            throw new InvalidContainerConfigurationProvidedException(
                    "Can't add a container with the same name as the current one");
        }

        final String addedContainerSubdomainName = addedContainer.getSubdomainName();
        if (addedContainerSubdomainName == null) {
            throw new InvalidContainerConfigurationProvidedException("The sub-domain the container '"
                    + addedContainerName + "' to add is not defined");
        } else if (!this.subDomainsConfig.containsKey(addedContainerSubdomainName)) {
            throw new SubdomainUnknownException(addedContainerSubdomainName);
        }

        synchronized(this.containersConfig) {
            if (this.containersConfig.containsKey(addedContainer.getName())) {
                throw new ContainerAlreadyExistException(addedContainerName);
            }
    
            // Register the new container
            this.containersConfig.put(addedContainerName, addedContainer);
        }
        
        for (final ContainerConfigurationListener addedContainerListener : this.containerConfigListeners) {
            addedContainerListener.onAddedContainerConfiguration(addedContainer);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void registerContainerConfigurationListener(final ContainerConfigurationListener listener) {
        this.containerConfigListeners.add(listener);
    }

    /**
     * {@inheritDoc}
     */
    public void deregisterContainerConfigurationListener(final ContainerConfigurationListener listener) {
        this.containerConfigListeners.remove(listener);
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void addSubdomainConfiguration(final SubDomainConfiguration addedSubdomain)
            throws SubdomainAlreadyExistException, NoSubdomainConfigurationProvidedException,
            InvalidSubdomainConfigurationProvidedException {

        if (addedSubdomain == null) {
            throw new NoSubdomainConfigurationProvidedException();
        }

        final String subdomainName = addedSubdomain.getName();
        if (subdomainName == null) {
            throw new InvalidSubdomainConfigurationProvidedException("The sub-domain name is not defined");
        } else if (subdomainName.equals(this.subDomainConfiguration.getName())) {
            throw new InvalidSubdomainConfigurationProvidedException(
                    "Can't add a sub-domain with the same name as the current one");
        }

        if (this.subDomainsConfig.containsKey(subdomainName)) {
            throw new SubdomainAlreadyExistException(subdomainName);
        } else {
            this.subDomainsConfig.put(subdomainName, addedSubdomain);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.kernel.configuration.ConfigurationService#
     * getContainerConfiguration()
     */
    public ContainerConfiguration getContainerConfiguration() {
        return this.containerConfiguration;
    }

    /**
     * {@inheritDoc}
     */
    public Map<String, String> getAllServerProperties() {
        final Map<String, String> containerLocalProperties = new HashMap<String, String>();

        containerLocalProperties.put(ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME,
                this.containerConfiguration.getDataRootPath());
        containerLocalProperties.put(ConfigurationProperties.REPOSITORY_PATH_PROPERTY_NAME,
                this.containerConfiguration.getRepositoryDirectoryPath());
        containerLocalProperties.put(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME,
                this.containerConfiguration.getWorkDirectoryPath());
        containerLocalProperties.put(ConfigurationProperties.EXCHANGE_VALIDATION,
                Boolean.toString(this.containerConfiguration.isExchangeValidation()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_QUEUE_MAXSIZE,
                Integer.toString(this.containerConfiguration.getTransporterQueueMaxSize()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_QUEUE_OFFERING_TIMEOUT,
                Long.toString(this.containerConfiguration.getTransporterQueueOfferingTimeout()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_TCP_RECEIVERS,
                Integer.toString(this.containerConfiguration.getTCPReceivers()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_TCP_SENDERS,
                Integer.toString(this.containerConfiguration.getTCPSenders()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_TCP_CONNECTION_TIMEOUT,
                Long.toString(this.containerConfiguration.getTCPConnectionTimeout()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_TCP_SEND_TIMEOUT,
                Long.toString(this.containerConfiguration.getTCPSendTimeout()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_TCP_SEND_EVICTOR_DELAY,
                Long.toString(this.containerConfiguration.getTCPSenderEvictorDelay()));
        containerLocalProperties.put(ConfigurationProperties.TRANSPORT_TCP_SEND_EVICTABLE_DELAY,
                Long.toString(this.containerConfiguration.getTCPSenderEvictableDelay()));
        containerLocalProperties.put(ConfigurationProperties.ROUTER_QOS,
                this.containerConfiguration.getRouterQOS());
        containerLocalProperties.put(ConfigurationProperties.ROUTER_STRATEGY,
                this.containerConfiguration.getRouterStrategy());
        containerLocalProperties.put(ConfigurationProperties.ROUTER_SEND_ATTEMPT,
                Short.toString(this.containerConfiguration.getRouterSendAttempt()));
        containerLocalProperties.put(ConfigurationProperties.ROUTER_SEND_DELAY,
                Integer.toString(this.containerConfiguration.getRouterSendDelay()));
        containerLocalProperties.put(ConfigurationProperties.TASK_TIMEOUT,
                Long.toString(this.containerConfiguration.getTaskTimeout()));
        containerLocalProperties.put(ConfigurationProperties.CLASSLOADER_ISOLATED,
                Boolean.toString(this.containerConfiguration.isIsolateJBIClassLoaders()));
        containerLocalProperties.put(ConfigurationProperties.SSL_KEY_PASSWORD,
                this.containerConfiguration.getSSLKeyPassword());
        containerLocalProperties.put(ConfigurationProperties.SSL_KEYSTORE_FILE,
                this.containerConfiguration.getSSLKeystore());
        containerLocalProperties.put(ConfigurationProperties.SSL_KEYSTORE_PASSWORD,
                this.containerConfiguration.getSSLKeystorePassword());
        containerLocalProperties.put(ConfigurationProperties.SSL_TRUSTSTORE_FILE,
                this.containerConfiguration.getSSLTruststore());
        containerLocalProperties.put(ConfigurationProperties.SSL_TRUSTSTORE_PASSWORD,
                this.containerConfiguration.getSSLTruststorePassword());
        containerLocalProperties.put(ConfigurationProperties.PERSISTENCE_DB_DRIVER,
                this.containerConfiguration.getPersistenceDatabaseDriver());
        containerLocalProperties.put(ConfigurationProperties.PERSISTENCE_DB_URL,
                this.containerConfiguration.getPersistenceDatabaseUrl());
        containerLocalProperties.put(ConfigurationProperties.PERSISTENCE_DB_USER,
                this.containerConfiguration.getPersistenceDatabaseUser());
        containerLocalProperties.put(ConfigurationProperties.PERSISTENCE_DB_PASSWORD,
                this.containerConfiguration.getPersistenceDatabasePassword());
        containerLocalProperties.put(ConfigurationProperties.PERSISTENCE_DURATION,
                Long.toString(this.containerConfiguration.getPersistenceDuration()));
        containerLocalProperties.put(ConfigurationProperties.PERSISTENCE_FETCHSIZE,
                Integer.toString(this.containerConfiguration.getPersistenceFetchSize()));
        containerLocalProperties.put(ConfigurationProperties.RECOVERY_CORE_POOL_SIZE,
                Integer.toString(this.containerConfiguration.getRecoveryCorePoolSize()));
        containerLocalProperties.put(ConfigurationProperties.RECOVERY_KEEP_ALIVE_TIME,
                Long.toString(this.containerConfiguration.getRecoveryKeepAliveTime()));
        containerLocalProperties.put(ConfigurationProperties.TOPOLOGY_DYNAMIC_LOCK_WAIT_TIME,
                Integer.toString(this.containerConfiguration.getTopologyDynamicLockWaitTime()));

        // Add user properties
        final Map<String, String> userConfiguration = this.containerConfiguration
                .getUserConfiguration();
        if (userConfiguration != null) {
            containerLocalProperties.putAll(userConfiguration);
        }

        // Add extra properties
        final Map<String, String> extraConfiguration = this.containerConfiguration
                .getExtraConfiguration();
        if (extraConfiguration != null) {
            containerLocalProperties.putAll(extraConfiguration);
        }

        return containerLocalProperties;
    }
    
    /**
     * {@inheritDoc}
     */
    public Properties getServerProperties() {
        final Properties serverProperties = new Properties();
        for (final Map.Entry<String, String> entry : this.getAllServerProperties().entrySet()) {
            final String value = entry.getValue();
            if (value != null) {
                serverProperties.setProperty(entry.getKey(), value);
            }
        }
        return serverProperties;
    }

    /**
     * {@inheritDoc}
     */
    public ContainerConfiguration getContainerConfiguration(final String containerName) {
        final ContainerConfiguration result;
        if (this.containerConfiguration.getName().equals(containerName)) {
            result = this.getContainerConfiguration();
        } else {
            result = this.containersConfig.get(containerName);
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized Set<ContainerConfiguration> getContainersConfiguration() {
        synchronized (this.containersConfig) {
            final Set<ContainerConfiguration> result = new HashSet<ContainerConfiguration>(this.containersConfig.size());
            for (final ContainerConfiguration containerCfg : this.containersConfig.values()) {
                result.add(containerCfg);
            }
            return result;
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized DomainConfiguration getDomainConfiguration() {
        this.log.call();
        return this.domainConfiguration;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized SubDomainConfiguration getSubDomainConfiguration() {
        this.log.call();
        return this.subDomainConfiguration;
    }

    /**
     * {@inheritDoc}
     */
    public SubDomainConfiguration getSubDomainConfiguration(final String subdomainName) {

        final SubDomainConfiguration result;
        if (subdomainName == null || this.subDomainConfiguration.getName().equals(subdomainName)) {
            result = this.getSubDomainConfiguration();
        } else {
            result = this.subDomainsConfig.get(subdomainName);
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized Set<SubDomainConfiguration> getSubDomainsConfiguration() {
        final Set<SubDomainConfiguration> result = new HashSet<SubDomainConfiguration>(
                this.subDomainsConfig.size());
        for (final SubDomainConfiguration subDomainCfg : this.subDomainsConfig.values()) {
            result.add(subDomainCfg);
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized Topology getInitialLocalTopology() throws ConfigurationException {
        this.log.call();

        return this.initialLocalTopology;
    }

    /**
     * {@inheritDoc}
     */
    public void removeContainerConfiguration(final ContainerConfiguration removedContainer)
            throws NoContainerConfigurationProvidedException, ContainerDoesNotExistException,
            InvalidContainerConfigurationProvidedException {
        if (removedContainer == null) {
            throw new NoContainerConfigurationProvidedException();
        }

        synchronized (this.containersConfig) {
            final String removedContainerName = removedContainer.getName();
            if (removedContainerName == null) {
                throw new ContainerDoesNotExistException(removedContainerName);
            } else if (removedContainerName.equals(this.containerConfiguration.getName())) {
                throw new InvalidContainerConfigurationProvidedException("Can't remove the current container");
            } else if (!this.containersConfig.containsKey(removedContainerName)) {
                throw new ContainerDoesNotExistException(removedContainerName);
            }

            this.containersConfig.remove(removedContainerName);
        }

        for (final ContainerConfigurationListener removedContainerListener : this.containerConfigListeners) {
            removedContainerListener.onRemovedContainerConfiguration(removedContainer);
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void removeSubdomainConfiguration(final SubDomainConfiguration removedSubdomain)
            throws NoSubdomainConfigurationProvidedException, SubdomainDoesNotExistException,
            InvalidSubdomainConfigurationProvidedException {
        if (removedSubdomain == null) {
            throw new NoSubdomainConfigurationProvidedException();
        }

        final String removedSubdomainName = removedSubdomain.getName();
        if (removedSubdomainName == null) {
            throw new SubdomainDoesNotExistException(removedSubdomainName);
        } else if (removedSubdomainName.equals(this.subDomainConfiguration.getName())) {
            throw new InvalidSubdomainConfigurationProvidedException("Can't remove the current sub-domain");
        } else if (!this.subDomainsConfig.containsKey(removedSubdomainName)) {
            throw new SubdomainDoesNotExistException(removedSubdomainName);
        }

        this.subDomainsConfig.remove(removedSubdomainName);
        this.removeSubdomainContainers(removedSubdomainName);
    }

    /**
     * Unregisters all the containers member of a sub-domain within the local configuration.
     * 
     * @param subdomainName
     *            The name of the sub-domain removed
     */
    private void removeSubdomainContainers(final String subdomainName) {

        synchronized (this.containersConfig) {
            final Iterator<Entry<String, ContainerConfiguration>> itContainerEntries = this.containersConfig.entrySet()
                    .iterator();
            while (itContainerEntries.hasNext()) {
                final Entry<String, ContainerConfiguration> containerEntry = itContainerEntries.next();
                if (containerEntry.getValue().getSubdomainName().equals(subdomainName)) {
                    itContainerEntries.remove();
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void setContainerState(ContainerState state) {
        this.containerConfiguration.setState(state);
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    @Lifecycle(step = Step.START)
    public void start() throws Exception {
        this.log.call();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.call();
    }

    /**
     * {@inheritDoc}
     */
    public void loadConfiguration(final Properties serverLocalProperties, final Topology topology)
            throws NoTopologyProvidedException, NoServerPropertiesProvidedException, MissingServerPropertyException,
            ConfigurationException {
        this.log.start();
        try {

            if (serverLocalProperties == null) {
                throw new NoServerPropertiesProvidedException();
            }

            if (topology == null) {
                throw new NoTopologyProvidedException();
            }

            this.initializeLocalProperties(this.containerConfiguration, serverLocalProperties);

            // Checks if the container name has been set in the
            // server local configuration (ie. content of 'server.properties'
            // or equivalent) or as a system property
            final String containerName = this.containerConfiguration.getName();
            if (containerName == null || containerName.isEmpty()) {
                throw new MissingServerPropertyException(
                        ConfigurationProperties.CONTAINER_NAME,
                        "The property '"
                                + ConfigurationProperties.CONTAINER_NAME
                                + "' must be set in the server local configuration (ie. in the content of file 'server.properties' or equivalent) or as system property");
            }

            // Check if a persisted topology definition exists
            final String workingAreaValue = this.containerConfiguration.getWorkDirectoryPath();
            final File workingArea = new File(workingAreaValue);
            if (workingArea.exists()) {
                if (workingArea.isDirectory()) {
                    final File persistedTopologyFile = new File(workingArea, PERSISTED_TOPOLOGY_DEFINITION);
                    if (persistedTopologyFile.exists()) {
                        if (persistedTopologyFile.isFile()) {
                            try {
                                this.log.debug("A persisted topology definition exists, we use it.");
                                this.initialLocalTopology = TopologyBuilder.createTopology(persistedTopologyFile);
                            } catch (final TopologyException e) {
                                throw new ConfigurationException(String.format(
                                        "An error occurs parsing the persisted topology definition '%s'.",
                                        persistedTopologyFile.getAbsolutePath()), e);
                            }
                        } else {
                            throw new ConfigurationException(String.format(
                                    "Invalid persisted topology definition '%s' (not a file).",
                                    persistedTopologyFile.getAbsolutePath()));
                        }
                    } else {
                        this.log.debug("No persisted topology definition exists.");
                        this.initialLocalTopology = topology;
                    }
                } else {
                    throw new ConfigurationException(String.format("Invalid working area '%s' (not a directory).",
                            workingAreaValue));
                }
            } else {
                this.log.info(String
                        .format("The working area '%s' does not exists. Probably it's the first startup of the container, or the data area has been cleaned manually.",
                                workingArea.getAbsoluteFile()));
                this.initialLocalTopology = topology;
            }

            this.validateTopology(this.containerConfiguration, this.initialLocalTopology);

            this.initializeDomainConfiguration(this.domainConfiguration, this.initialLocalTopology);
            this.initializeSubDomainConfiguration(this.containerConfiguration, this.subDomainConfiguration,
                    this.initialLocalTopology);
            this.initializeContainerConfiguration(this.containerConfiguration, this.initialLocalTopology);

            this.registerSubdomains(this.subDomainsConfig, this.containersConfig, this.initialLocalTopology);

            // post validate platform configuration
            this.postValidateContainerConfiguration(this.containerConfiguration);

            if (this.log.isDebugEnabled()) {
                this.log.debug("Configuration used :");
                this.log.debug(this.containerConfiguration);
            }

        } catch (final ConfigurationException e) {
            this.log.error("Failed to load configuration", e);
            throw e;
        } finally {
            this.log.end();
        }
    }

    /**
     * <p>
     * Registers the containers defined in the domain.
     * <p>
     * <p>
     * If a container is already registered but not declared in the topology, it is unregistered
     * </p>
     * 
     * @param containersConfig
     *            The list of registered container configurations. Can not be {@code null}.
     * @param topology
     *            A topology configuration coming from its XML definition. Can not be <code>null</code>.
     */
    private void registerContainers(final Map<String, ContainerConfiguration> containersConfig, final Topology topology) {
        final Collection<Container> containers = TopologyHelper.findAllContainers(topology);
        for (final Container container : containers) {
            final ContainerConfiguration cc = new ContainerConfiguration();
            ConfigurationServiceUtils.initializeContainerConfiguration(cc, container,
                    TopologyHelper.findSubdomain(container, topology).getName());
            containersConfig.put(cc.getName(), cc);
        }

        final Iterator<ContainerConfiguration> itContainerCfg = containersConfig.values().iterator();
        while (itContainerCfg.hasNext()) {
            final ContainerConfiguration containerCfg = itContainerCfg.next();
            boolean bFind = false;
            for (final Container container : containers) {
                if (containerCfg.getName().equals(container.getName())) {
                    bFind = true;
                    break;
                }
            }
            if (!bFind) {
                // The container is no more in the topology
                itContainerCfg.remove();
            }
        }
    }

    /**
     * <p>
     * Registers the sub-domains defined in the domain.
     * </p>
     * <p>
     * Note:
     * <ul>
     * <li>When registering a sub-domain, its containers are also automatically registered in
     * <code>containersConfig</code></li>
     * <li>
     * If a sub-domain is already registered but not declared in the topology, it is unregistered and its containers are
     * also automatically unregistered.</li>
     * </p>
     * 
     * @param subdomainsConfig
     *            The list of registered sub-domain configurations. Can not be {@code null}.
     * @param containersConfig
     *            The list of registered container configurations. Can not be {@code null}.
     * @param topology
     *            A topology configuration coming from its XML definition. Can not be <code>null</code>.
     * @throws ConfigurationException
     *             Something is wrong in the configuration of a sub-domain to register
     */
    private void registerSubdomains(final Map<String, SubDomainConfiguration> subdomainsConfig,
            final Map<String, ContainerConfiguration> containersConfig, final Topology topology)
            throws ConfigurationException {
        final List<Subdomain> subdomains = topology.getDomain().getSubDomain();
        for (final Subdomain subdomain : subdomains) {
            final SubDomainConfiguration subDomainCfg = new SubDomainConfiguration();

            this.initializeSubDomainConfiguration(subDomainCfg, subdomain);
            this.subDomainsConfig.put(subdomain.getName(), subDomainCfg);

            this.registerContainers(containersConfig, topology);
        }

        final Iterator<SubDomainConfiguration> itSubdomainCfg = subdomainsConfig.values().iterator();
        while (itSubdomainCfg.hasNext()) {
            final SubDomainConfiguration subdomainCfg = itSubdomainCfg.next();
            boolean bFind = false;
            for (final Subdomain subdomain : subdomains) {
                if (subdomainCfg.getName().equals(subdomain.getName())) {
                    bFind = true;
                    break;
                }
            }
            if (!bFind) {
                // The sub-domain is no more in the topology
                itSubdomainCfg.remove();

                // Unregister containers of the sub-domain
                final Iterator<ContainerConfiguration> itContainersConfig = containersConfig.values().iterator();
                while (itContainersConfig.hasNext()) {
                    final ContainerConfiguration containerConfig = itContainersConfig.next();
                    if (containerConfig.getSubdomainName().equals(subdomainCfg.getName())) {
                        // Container member of the removed sub-domain
                        itContainersConfig.remove();
                    }
                }
            }
        }
    }

    /**
     * Initializes a domain configuration from a topology JAXB element.
     * 
     * @param domainConfiguration
     *            The domain configuration to initialize from the topology <code>topology</code>. Can not be
     *            {@code null}.
     * @param topology
     *            A topology configuration coming from its XML definition to use to initialize the domain configuration
     *            <code>domainConfiguration</code>. Can not be <code>null</code>.
     * @throws ConfigurationException
     *             Something is missing
     */
    private void initializeDomainConfiguration(final DomainConfiguration domainConfiguration, final Topology topology)
            throws ConfigurationException {
        this.log.start();

        try {
            // domain mode
            domainConfiguration.setName(topology.getDomain().getName());
            domainConfiguration.setDescription(topology.getDomain().getDescription());

            // domain JNDI
            if (topology.getDomain().getJndi() != null) {
                final JndiConfiguration jndiConfiguration = new JndiConfiguration();

                jndiConfiguration.setJndiFactory(topology.getDomain().getJndi().getFactory());
                try {
                    jndiConfiguration.setJndiProviderUrl(new URI(topology.getDomain().getJndi().getProviderUrl()));
                } catch (URISyntaxException e) {
                    throw new ConfigurationException(e);
                }
                jndiConfiguration.setJndiSecurityPrincipal(topology.getDomain().getJndi().getSecurityPrincipal());
                jndiConfiguration.setJndiSecurityCredentials(topology.getDomain().getJndi()
                        .getSecurityCredentials());

                final Integer poolSize = topology.getDomain().getJndi().getPoolSize();
                if (poolSize != null) {
                    jndiConfiguration.setJndiPoolSize(poolSize.intValue());
                }
                final Integer batchSize = topology.getDomain().getJndi().getBatchSize();
                if (batchSize != null) {
                    jndiConfiguration.setJndiBatchSize(batchSize.intValue());
                }

                domainConfiguration.setJndiConfiguration(jndiConfiguration);
            }
        } finally {
            this.log.end();
        }
    }

    /**
     * Initialize a container configuration with the server properties.
     * 
     * @param containerConfiguration
     *            The container configuration to initialize. Can be <code>null</code>.
     * @param serverLocalProperties
     *            The local properties of the current container (ie: the content of 'server.properties' or equivalent).
     *            Can be <code>null</code>.
     * @throws MissingServerPropertyException
     *             A required server property is missing.
     * @throws ConfigurationException
     *             An error occurs about a property value
     */
    private void initializeLocalProperties(final ContainerConfiguration containerConfiguration,
            final Properties serverLocalProperties) throws MissingServerPropertyException, ConfigurationException {
        this.log.start();

        final Map<String, String> userProperties = new HashMap<String, String>();

        final Map<String, String> extraProperties = new HashMap<String, String>();

        if (serverLocalProperties != null) {
            for (final Map.Entry<Object, Object> entry : serverLocalProperties.entrySet()) {
                final Object keyObj = entry.getKey();
                final String value = (String) entry.getValue();
                if (keyObj instanceof String) {
                    final String key = (String) keyObj;
                    if (key.equals(ConfigurationProperties.CONTAINER_NAME)) {
                        containerConfiguration.setName(value);
                    } else if (key
                            .equals(ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME)) {
                        containerConfiguration.setDataRootPath(value);
                    } else if (key.equals(ConfigurationProperties.REPOSITORY_PATH_PROPERTY_NAME)) {
                        containerConfiguration.setRepositoryDirectoryPath(value);
                    } else if (key.equals(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME)) {
                        containerConfiguration.setWorkDirectoryPath(value);
                    } else if (key.equals(ConfigurationProperties.EXCHANGE_VALIDATION)) {
                        containerConfiguration.setExchangeValidation(Boolean
                                .parseBoolean(value));
                    } else if (key.equals(ConfigurationProperties.TRANSPORT_QUEUE_MAXSIZE)) {
                        containerConfiguration.setTransporterQueueMaxSize(Integer
                                .parseInt(value));
                    } else if (key.equals(ConfigurationProperties.TRANSPORT_QUEUE_OFFERING_TIMEOUT)) {
                        containerConfiguration.setTransporterQueueOfferingTimeout(Long
                                .parseLong(value));
                    } else if (key.equals(ConfigurationProperties.TRANSPORT_TCP_RECEIVERS)) {
                        containerConfiguration.setTCPReceivers(Integer.parseInt(value));
                    } else if (key.equals(ConfigurationProperties.TRANSPORT_TCP_SENDERS)) {
                        containerConfiguration.setTCPSenders(Integer.parseInt(value));
                    } else if (key.equals(ConfigurationProperties.TRANSPORT_TCP_CONNECTION_TIMEOUT)) {
                        containerConfiguration.setTCPConnectionTimeout(Long.parseLong(value));
                    } else if (key.equals(ConfigurationProperties.TRANSPORT_TCP_SEND_TIMEOUT)) {
                        containerConfiguration.setTCPSendTimeout(Long.parseLong(value));
                    } else if (key.equals(ConfigurationProperties.TRANSPORT_TCP_SEND_EVICTOR_DELAY)) {
                        containerConfiguration.setTCPSenderEvictorDelay(Long.parseLong(value));
                    } else if (key
                            .equals(ConfigurationProperties.TRANSPORT_TCP_SEND_EVICTABLE_DELAY)) {
                        containerConfiguration.setTCPSenderEvictableDelay(Long
                                .parseLong(value));
                    } else if (key.equals(ConfigurationProperties.ROUTER_QOS)) {
                        containerConfiguration.setRouterQOS(value);
                    } else if (key.equals(ConfigurationProperties.ROUTER_STRATEGY)) {
                        containerConfiguration.setRouterStrategy(value);
                    } else if (key.equals(ConfigurationProperties.ROUTER_SEND_ATTEMPT)) {
                        containerConfiguration.setRouterSendAttempt(Short.parseShort(value));
                    } else if (key.equals(ConfigurationProperties.ROUTER_SEND_DELAY)) {
                        containerConfiguration.setRouterSendDelay(Integer.parseInt(value));
                    } else if (key.equals(ConfigurationProperties.TASK_TIMEOUT)) {
                        containerConfiguration.setTaskTimeout(Long.parseLong(value));
                    } else if (key.equals(ConfigurationProperties.CLASSLOADER_ISOLATED)) {
                        containerConfiguration.setIsolateJBIClassLoaders(Boolean
                                .parseBoolean(value));
                    } else if (key.equals(ConfigurationProperties.SSL_KEY_PASSWORD)) {
                        containerConfiguration.setSSLKeyPassword(value);
                    } else if (key.equals(ConfigurationProperties.SSL_KEYSTORE_FILE)) {
                        containerConfiguration.setSSLKeystore(value);
                    } else if (key.equals(ConfigurationProperties.SSL_KEYSTORE_PASSWORD)) {
                        containerConfiguration.setSSLKeystorePassword(value);
                    } else if (key.equals(ConfigurationProperties.SSL_TRUSTSTORE_FILE)) {
                        containerConfiguration.setSSLTruststore(value);
                    } else if (key.equals(ConfigurationProperties.SSL_TRUSTSTORE_PASSWORD)) {
                        containerConfiguration.setSSLTruststorePassword(value);
                    } else if (key.equals(ConfigurationProperties.PERSISTENCE_DB_DRIVER)) {
                        containerConfiguration.setPersistenceDatabaseDriver(value);
                    } else if (key.equals(ConfigurationProperties.PERSISTENCE_DB_URL)) {
                        containerConfiguration.setPersistenceDatabaseUrl(value);
                    } else if (key.equals(ConfigurationProperties.PERSISTENCE_DB_USER)) {
                        containerConfiguration.setPersistenceDatabaseUser(value);
                    } else if (key.equals(ConfigurationProperties.PERSISTENCE_DB_PASSWORD)) {
                        containerConfiguration.setPersistenceDatabasePassword(value);
                    } else if (key.equals(ConfigurationProperties.PERSISTENCE_DURATION)) {
                        containerConfiguration.setPersistenceDuration(Long.parseLong(value));
                    } else if (key.equals(ConfigurationProperties.PERSISTENCE_FETCHSIZE)) {
                        containerConfiguration
                                .setPersistenceFetchSize(Integer.parseInt(value));
                    } else if (key.equals(ConfigurationProperties.TOPOLOGY_PASSPHRASE)) {
                        containerConfiguration.setTopologyPassPhrase(serverLocalProperties
                                .getProperty(key, null));
                    } else if (key.equals(ConfigurationProperties.TOPOLOGY_DYNAMIC_LOCK_WAIT_TIME)) {
                        containerConfiguration.setTopologyDynamicLockWaitTime(Integer.parseInt(value));
                    } else if (key.equals(ConfigurationProperties.TOPOLOGY_PINGER_START_DELAY)) {
                        containerConfiguration.setTopologyPingerStartDelay(Long.parseLong(value));
                    } else if (key.equals(ConfigurationProperties.TOPOLOGY_PINGER_PERIOD_DELAY)) {
                        containerConfiguration.setTopologyPingerPeriodDelay(Long.parseLong(value));
                    } else if (key.startsWith(ConfigurationProperties.USER_CONFIG_PREFIX + ".")) {
                        userProperties.put(key, value);
                    } else if (key.equals(ConfigurationProperties.RECOVERY_CORE_POOL_SIZE)) {
                        containerConfiguration
                                .setRecoveryCorePoolSize(Integer.parseInt(value));
                    } else if (key.equals(ConfigurationProperties.RECOVERY_KEEP_ALIVE_TIME)) {
                        containerConfiguration.setRecoveryKeepAliveTime(Long.parseLong(value));
                    } else {
                        extraProperties.put(key, value);
                    }
                }
            }
        }

        // If the container name is not defined as server property, try to get it as system property
        if (this.containerConfiguration.getName() == null) {
            this.containerConfiguration.setName(System.getProperty(ConfigurationProperties.CONTAINER_NAME));
        }

        // Set the container name into the Petals MDC as soon as possible
        if (this.containerConfiguration.getName() != null) {
            PetalsExecutionContext.putContainerName(this.containerConfiguration.getName());
        }

        if (userProperties.size() > 0) {
            containerConfiguration.setUserConfiguration(userProperties);
        }

        containerConfiguration.setExtraConfiguration(extraProperties);

        // Set default value to not set required properties
        try {
            if (containerConfiguration.getRepositoryDirectoryPath() == null) {
                final String dataRootPath = containerConfiguration.getDataRootPath();
                if (dataRootPath != null && !dataRootPath.isEmpty()) {
                    containerConfiguration.setRepositoryDirectoryPath(PropertiesHelper
                            .resolveString(ConfigurationProperties.DEFAULT_REPOSITORY_PATH,
                                    serverLocalProperties));
                }
                else {
                    throw new MissingServerPropertyException(ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME);
                }
            }

            if (containerConfiguration.getWorkDirectoryPath() == null) {
                final String dataRootPath = containerConfiguration.getDataRootPath();
                if (dataRootPath != null && !dataRootPath.isEmpty()) {
                    containerConfiguration.setWorkDirectoryPath(PropertiesHelper.resolveString(
                            ConfigurationProperties.DEFAULT_WORKING_AREA_PATH, serverLocalProperties));
                }
                else {
                    throw new MissingServerPropertyException(ConfigurationProperties.DATA_ROOT_DIRECTORY_PROPERTY_NAME);
                }
            }

            // TODO: Should be move in the right fractal component startup when
            // Persistence service and Registry service will use different
            // parameters
            if (containerConfiguration.getPersistenceDatabaseUrl() == null) {
                final String workingRootPath = containerConfiguration.getWorkDirectoryPath();
                if (workingRootPath != null && !workingRootPath.isEmpty()) {
                    containerConfiguration.setPersistenceDatabaseUrl(PersistenceService.DEFAULT_URL_PREFIX
                            + containerConfiguration.getWorkDirectoryPath() + File.separator
                            + PersistenceService.DEFAULT_URL_SUFFIX);
                } else {
                    throw new MissingServerPropertyException(ConfigurationProperties.WORKING_AREA_PATH_PROPERTY_NAME);
                }
            }

        } catch (final PropertiesException e) {
            throw new ConfigurationException(e);
        }

        this.log.end();
    }

    /**
     * <p>
     * Initializes a container configuration bean from a topology JAXB element.
     * </p>
     * <p>
     * To be able to retrieve the container into the topology, its name <b>MUST<b> initialized in
     * <code>containerConfiguration</code>.
     * </p>
     * 
     * @param containerConfiguration
     *            The container configuration to initialize from the topology <code>topology</code>. Can not be
     *            {@code null}.
     * @param topology
     *            A topology configuration coming from its XML definition to use to initialize the container
     *            configuration <code>containerConfiguration</code>. Can not be <code>null</code>.
     * @throws ConfigurationException
     *             Something is missing
     */
    private void initializeContainerConfiguration(final ContainerConfiguration containerConfiguration,
            final Topology topology) throws ConfigurationException {
        this.log.start();

        try {
            final String containerName = containerConfiguration.getName();
            final Container localContainer = TopologyHelper.findContainer(containerName, topology);
            if (localContainer == null) {
                throw new ConfigurationException(
                        "Failed to retrieve the container configuration for the container named '" + containerName
                                + "'");
            }

            ConfigurationServiceUtils.initializeContainerConfiguration(containerConfiguration, localContainer,
                    TopologyHelper.findSubdomain(localContainer, topology).getName());
        } finally {
            this.log.end();
        }
    }

    /**
     * Initializes a sub-domain configuration from a topology JAXB element.
     * 
     * @param containerConfiguration
     *            The container configuration associated to the Petals node for which the sub-domain must be
     *            initialized. Can not be {@code null}.
     * @param subdomainConfiguration
     *            The sub-domain configuration to initialize from the topology <code>topology</code>. Can not be
     *            {@code null}.
     * @param topology
     *            A topology configuration coming from its XML definition to use to initialize the sub-domain
     *            configuration <code>subdomainConfiguration</code>. Can not be <code>null</code>.
     * @throws ContainerDoesNotExistException
     *             The container does not exist in the topology
     * @throws ConfigurationException
     *             Something is missing
     */
    private void initializeSubDomainConfiguration(final ContainerConfiguration containerConfiguration,
            final SubDomainConfiguration subdomainConfiguration, final Topology topology)
            throws ContainerDoesNotExistException, ConfigurationException {
        this.log.start();

        try {
            final Container container = TopologyHelper.findContainer(containerConfiguration.getName(), topology);
            if (container == null) {
                throw new ContainerDoesNotExistException(containerConfiguration.getName());
            }
            final Subdomain subDomain = TopologyHelper.findSubdomain(container, topology);
            if (subDomain == null) {
                throw new ConfigurationException(
                        "Failed to retrieve the sub-domain configuration for the sub-domain containing the container named '"
                                + container.getName() + "'");
            }

            this.initializeSubDomainConfiguration(subdomainConfiguration, subDomain);
        } finally {
            this.log.end();
        }
    }

    /**
     * Initializes a sub-domain configuration from a sub-domain JAXB element.
     * 
     * @param subdomainConfiguration
     *            The sub-domain configuration to initialize from the sub-domain <code>subDomain</code>. Can not be
     *            <code>null</code>.
     * @param subDomain
     *            A sub-domain configuration coming from its XML definition to use to initialize the sub-domain
     *            configuration <code>subdomainConfiguration</code>. Can not be <code>null</code>.
     * @throws ConfigurationException
     *             Something is wrong in the sub-domain configuration
     */
    private void initializeSubDomainConfiguration(final SubDomainConfiguration subdomainConfiguration,
            final Subdomain subDomain) throws ConfigurationException {
        this.log.start();

        try {
            subdomainConfiguration.setName(subDomain.getName());
            subdomainConfiguration.setDescription(subDomain.getDescription());
            if (SubdomainMode.STATIC.equals(subDomain.getMode())) {
                subdomainConfiguration
                        .setMode(org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration.SubdomainMode.STATIC);
            } else if (SubdomainMode.DYNAMIC.equals(subDomain.getMode())) {
                subdomainConfiguration
                        .setMode(org.ow2.petals.microkernel.api.configuration.SubDomainConfiguration.SubdomainMode.DYNAMIC);
            } else {
                throw new ConfigurationException("Unknown sub-domain mode");
            }
            if (subDomain.isNetworkTimeSynchronized() != null) {
                subdomainConfiguration.setNetworkTimeSynchronized(subDomain.isNetworkTimeSynchronized().booleanValue());
            }
            if (subDomain.getRegistryImplementation() != null) {
                subdomainConfiguration.setRegistryImplementation(subDomain.getRegistryImplementation().trim());
            }
            if (subDomain.getAny() != null) {
                subdomainConfiguration.setExtraParameters(subDomain.getAny());
            }
        } finally {
            this.log.end();
        }
    }

    /**
     * <p>
     * Post validation of a container configuration, based on {@link ContainerConfiguration} content:
     * <ul>
     * <li>check the coherence of the container configuration,</li>
     * <li>create required directories if needed.</li>
     * </ul>
     * </p>
     * 
     * @throws ConfigurationException
     *             Something is wrong
     */
    private void postValidateContainerConfiguration(final ContainerConfiguration containerConfiguration)
            throws ConfigurationException {

        // validate the routing policy and transporter requirements
        if (RouterService.FAST_POLICY.equals(containerConfiguration.getRouterQOS())
                && (containerConfiguration.getTCPPort() == 0)) {
            throw new ConfigurationException("The router transport QOS '" + RouterService.FAST_POLICY
                    + "' is not compatible with the transport protocols configuration");
        }
        if (RouterService.SECURE_POLICY.equals(containerConfiguration.getRouterQOS())) {
            throw new ConfigurationException("The default router QOS '" + RouterService.SECURE_POLICY
                    + "' is not compatible with the transport protocols configuration");
        }

        final String workingAreaValue = containerConfiguration.getWorkDirectoryPath();
        final File workingArea = new File(workingAreaValue);
        if (!workingArea.exists()) {
            if (!workingArea.mkdirs()) {
                this.log.warning(String
                        .format("Unable to create the working area '%s'. Potential problems can occurs later.",
                                workingAreaValue));
            }
        } else if (!workingArea.isDirectory()) {
            throw new ConfigurationException(String.format(
                    "Invalid working area '%s' (not a directory).", workingAreaValue));
        }
    }

    /**
     * Validate the consistency of the topology against a container configuration.
     * 
     * @param containerConfiguration
     *            A container configuration to validate against the topology <code>topology</code>. Can not be
     *            {@code null}
     * @param topology
     *            The topology to validate against the container configuration <code>containerConfiguration</code>. Can
     *            not be <code>null</code>.
     * @throws InconsistentConfigurationException
     *             The configuration is inconsistent.
     */
    private void validateTopology(final ContainerConfiguration containerConfiguration, final Topology topology)
            throws InconsistentConfigurationException {
        this.log.start();

        try {

            // Checks if the current container name can be found in the topology
            final String containerName = containerConfiguration.getName();
            if (TopologyHelper.findContainer(containerName, topology) == null) {
                throw new InconsistentContainerNameException(containerName);
            }
            // the container name has been found in the topology

            // validate JNDI consistency
            final Domain domain = topology.getDomain();
            final Jndi jndi = domain.getJndi();
            if (jndi != null) {
                final String jndiFactory = jndi.getFactory();
                if (jndiFactory == null || jndiFactory.trim().isEmpty()) {
                    throw new InconsistentDomainJndiFactoryException();
                }
                if (jndi.getProviderUrl() == null) {
                    throw new InconsistentDomainJndiProviderUrlException();
                }
            }

            // validate sub-domains
            final Map<String, Set<Integer>> networkPortAlreaySet = new HashMap<String, Set<Integer>>();
            this.validateSubdomains(domain.getSubDomain(), networkPortAlreaySet);

        } finally {
            this.log.end();
        }
    }

    /**
     * Validate the consistency of the topology sub-domains
     * 
     * @param subdomains
     *            The topology sub-domains to validate. Can not be <code>null</code>.
     * @param networkPortAlready
     *            List of the network ports. This list is completed during validation to detect duplicated ports.
     * @throws InconsistentConfigurationException
     *             The configuration is inconsistent.
     */
    private void validateSubdomains(final List<Subdomain> subdomains,
            final Map<String, Set<Integer>> networkPortAlreadySet) throws InconsistentConfigurationException {
        this.log.start();

        try {

            final Collection<String> subdomainNames = new HashSet<String>();
            for (final Subdomain subdomain : subdomains) {

                // Check the sub-domain name
                if (!subdomainNames.add(subdomain.getName())) {
                    throw new SubdomainNameDuplicatedException(subdomain.getName());
                }

                // Validate the domain mode
                final SubdomainMode subdomainMode = subdomain.getMode();
                if (!subdomainMode.equals(SubdomainMode.STATIC) && !subdomainMode.equals(SubdomainMode.DYNAMIC)) {
                    throw new InconsistentSubdomainModeException(subdomain.getName());
                }

                // validate containers
                this.validateContainers(subdomain.getContainer(), networkPortAlreadySet, subdomains.size() > 1);

            }
        } finally {
            this.log.end();
        }
    }

    /**
     * Validate the consistency of the topology containers
     * 
     * @param containers
     *            The topology containers to validate. Can not be <code>null</code>.
     * @param networkPortAlready
     *            List of the network ports. This list is completed during validation to detect duplicated ports.
     * @param isTransportServiceRequired
     *            The transport service is required at container level.
     * @throws InconsistentConfigurationException
     *             The configuration is inconsistent.
     */
    private void validateContainers(final List<Container> containers,
            final Map<String, Set<Integer>> networkPortAlreaySet, final boolean isTransportServiceRequired)
            throws InconsistentConfigurationException {
        this.log.start();

        try {
            final Collection<String> containerNames = new HashSet<String>();
            for (final Container container : containers) {

                if (!containerNames.add(container.getName())) {
                    throw new ContainerNameDuplicatedException(container.getName());
                }

                final String containerHost = container.getHost();
                if (containerHost == null) {
                    throw new InconsistentContainerHostException(container.getName());
                }

                if ((isTransportServiceRequired || containers.size() > 1) && container.getTransportService() == null) {
                    throw new InconsistentContainerTransportSvcException(container.getName());
                }

                // Validate network ports
                final Set<Integer> containerNetworkPortAlreaySet;
                if (!networkPortAlreaySet.containsKey(containerHost)) {
                    containerNetworkPortAlreaySet = new HashSet<Integer>();
                    networkPortAlreaySet.put(containerHost, containerNetworkPortAlreaySet);
                } else {
                    containerNetworkPortAlreaySet = networkPortAlreaySet.get(containerHost);
                }

                if (container.getJmxService().getRmiPort() != 0) {
                    if (!containerNetworkPortAlreaySet.add(container.getJmxService().getRmiPort())) {
                        throw new InconsistentContainerDuplicatedPortException(container.getName(), container
                                .getJmxService().getRmiPort());
                    }
                }

                if ((container.getTransportService() != null) && (container.getTransportService().getTcpPort() >= 0)) {
                    if (!containerNetworkPortAlreaySet.add(container.getTransportService().getTcpPort())) {
                        throw new InconsistentContainerDuplicatedPortException(container.getName(), container
                                .getTransportService().getTcpPort());
                    }
                }

            }
        } finally {
            this.log.end();
        }
    }

    /**
     * Verify the given new container configuration against the existing configuration.
     * 
     * @param newContainerConfiguration
     *            The new container configuration
     * @throws InconsistentContainerConfigurationException
     *             Something is incompatible between the existing container configuration and the new one
     */
    private void verifyContainer(final ContainerConfiguration newContainerConfiguration)
            throws InconsistentContainerConfigurationException {

        this.log.start();
        try {
            // If the new container already exists, the configurations must match
            ContainerConfiguration containerConfig = null;
            for (final ContainerConfiguration container : this.containersConfig.values()) {
                if (container.getName().equals(newContainerConfiguration.getName())) {
                    containerConfig = container;
                }
            }
            if (containerConfig != null) {
                if (!newContainerConfiguration.getHost().equals(this.containerConfiguration.getHost())) {
                    throw new InconsistentContainerHostException(newContainerConfiguration.getName());
                }

                if (!StringHelper.equal(containerConfig.getUser(), newContainerConfiguration.getUser())) {
                    throw new InconsistentContainerUserException(newContainerConfiguration.getName());
                }
                if (!StringHelper.equal(containerConfig.getPassword(), newContainerConfiguration.getPassword())) {
                    throw new InconsistentContainerPasswordException(newContainerConfiguration.getName());
                }

                // JMX Service
                if (containerConfig.getJmxRMIConnectorPort() != newContainerConfiguration.getJmxRMIConnectorPort()) {
                    throw new InconsistentContainerJmxPortException(newContainerConfiguration.getName());
                }

                // Transport service
                if (containerConfig.getTCPPort() != newContainerConfiguration.getTCPPort()) {
                    throw new InconsistentContainerFastRemoteTcpPortException(newContainerConfiguration.getName());
                }
            }
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void updateTopology(final Topology topology) throws NoTopologyProvidedException, ConfigurationException {

        if (topology == null) {
            throw new NoTopologyProvidedException();
        }

        // Check that the domain configuration of the new topology is compliant with the existing domain
        final DomainConfiguration newDomainConfiguration = new DomainConfiguration();
        this.initializeDomainConfiguration(newDomainConfiguration, topology);
        this.verifyDomainConfiguration(newDomainConfiguration);

        // Check that the current sub-domain already exists
        final Subdomain subdomain = TopologyHelper.findSubdomain(this.subDomainConfiguration.getName(), topology);
        if (subdomain == null) {
            throw new InconsistentCurrentSubdomainNoMoreExistException();
        }

        // Check that the sub-domain configuration of the new topology associated to the current container is
        // compliant with the existing sub-domain
        final SubDomainConfiguration newSubdomainConfiguration = new SubDomainConfiguration();
        try {
            this.initializeSubDomainConfiguration(this.containerConfiguration, newSubdomainConfiguration, topology);
        } catch (final ContainerDoesNotExistException e) {
            // The current container is not defined in the new topology
            throw new InconsistentCurrentContainerNoMoreExistException();
        }
        this.verifySubDomain(newSubdomainConfiguration);

        // Check that the sub-domain contains the current container
        if (TopologyHelper.findContainer(this.containerConfiguration.getName(), subdomain) == null) {
            throw new InconsistentCurrentContainerMovedException();
        }

        // Check that the current container configuration of the new topology is compliant with the existing one
        final ContainerConfiguration newContainerConfiguration = new ContainerConfiguration();
        newContainerConfiguration.setName(this.containerConfiguration.getName());
        this.initializeContainerConfiguration(newContainerConfiguration, topology);
        this.verifyContainer(newContainerConfiguration);

        // validate the consistency of the new topology with the current container
        this.validateTopology(this.containerConfiguration, topology);

        // The dynamic topology is compliant with the one already known
        // we update the whole configuration
        this.registerSubdomains(this.subDomainsConfig, this.containersConfig, topology);

    }

    /**
     * Verify the given new domain configuration against the existing configuration.
     * 
     * @param newDomainConfiguration
     *            The new domain configuration
     * @throws InconsistentDomainConfigurationException
     *             Something is incompatible between the existing domain configuration and the new one
     */
    private void verifyDomainConfiguration(final DomainConfiguration newDomainConfiguration)
            throws InconsistentDomainConfigurationException {
        this.log.start();
        try {

            // JNDI service
            final JndiConfiguration jndiConfiguration = this.domainConfiguration.getJndiConfiguration();
            final JndiConfiguration newJndiConfiguration = newDomainConfiguration.getJndiConfiguration();
            if ((jndiConfiguration == null && newJndiConfiguration != null)
                    || (jndiConfiguration != null && newJndiConfiguration == null)) {
                throw new InconsistentDomainConfigurationException("JNDI configuration mismatch");
            } else if (jndiConfiguration != null && newJndiConfiguration != null) {
                if (!StringHelper.equal(newJndiConfiguration.getJndiFactory(), jndiConfiguration.getJndiFactory())) {
                    throw new InconsistentDomainJndiFactoryException();
                }
                if ((newJndiConfiguration.getJndiProviderUrl() == null && jndiConfiguration.getJndiProviderUrl() != null)
                        || (newJndiConfiguration.getJndiProviderUrl() != null)
                        && (jndiConfiguration.getJndiProviderUrl() == null)
                        || (newJndiConfiguration.getJndiProviderUrl() != null
                                && jndiConfiguration.getJndiProviderUrl() != null && !newJndiConfiguration
                                .getJndiProviderUrl().equals(jndiConfiguration.getJndiProviderUrl()))) {
                    throw new InconsistentDomainJndiProviderUrlException();
                }
                if (!StringHelper.equal(newJndiConfiguration.getJndiSecurityPrincipal(),
                        jndiConfiguration.getJndiSecurityPrincipal())) {
                    throw new InconsistentDomainConfigurationException("JNDI Security Principal mismatch");
                }
                if (!StringHelper.equal(newJndiConfiguration.getJndiSecurityCredentials(),
                        jndiConfiguration.getJndiSecurityCredentials())) {
                    throw new InconsistentDomainConfigurationException("JNDI Security Credentials mismatch");
                }
                if (newJndiConfiguration.getJndiPoolSize() != jndiConfiguration.getJndiPoolSize()) {
                    throw new InconsistentDomainConfigurationException("JNDI Pool Size mismatch");
                }
                if (newJndiConfiguration.getJndiBatchSize() != jndiConfiguration.getJndiBatchSize()) {
                    throw new InconsistentDomainConfigurationException("JNDI Batch Size mismatch");
                }
            }

        } finally {
            this.log.end();
        }
    }

    /**
     * Verify the given new sub-domain configuration against the existing sub-domain configuration.
     * 
     * @param newSubdomainConfiguration
     *            The new sub-domain configuration
     * @throws InconsistentSubdomainConfigurationException
     *             Something is incompatible between the existing sub-domain configuration and the new one
     */
    private void verifySubDomain(final SubDomainConfiguration newSubdomainConfiguration)
            throws InconsistentSubdomainConfigurationException {
        this.log.start();

        try {
            // If the sub-domain already exists, the configurations must match
            SubDomainConfiguration subdomainConfig = null;
            for (final SubDomainConfiguration subdomain : this.subDomainsConfig.values()) {
                if (subdomain.getName().equals(newSubdomainConfiguration.getName())) {
                    subdomainConfig = subdomain;
                }

                if (!newSubdomainConfiguration.getMode().equals(this.subDomainConfiguration.getMode())) {
                    throw new InconsistentSubdomainModeException(newSubdomainConfiguration.getName());
                }
            }
            if (subdomainConfig != null) {
                if (newSubdomainConfiguration.isNetworkTimeSynchronized() != subdomainConfig
                        .isNetworkTimeSynchronized()) {
                    throw new InconsistentNetworkTimeSyncException(subdomainConfig.isNetworkTimeSynchronized(),
                            newSubdomainConfiguration.isNetworkTimeSynchronized());
                }

            }
        } finally {
            this.log.end();
        }
    }

    /**
     * {@inheritDoc}
     */
    public void reconfigureDomain(final DomainConfiguration newDomainConfiguration)
            throws NoDomainConfigurationProvidedException, NoDomainNameProvidedException {

        if (newDomainConfiguration == null) {
            throw new NoDomainConfigurationProvidedException();
        } else {
            final String newDomainName = newDomainConfiguration.getName();
            if (newDomainName != null && !newDomainName.isEmpty()) {
                this.domainConfiguration.setName(newDomainName);
                this.domainConfiguration.setDescription(newDomainConfiguration.getDescription());
            } else {
                throw new NoDomainNameProvidedException();
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void reconfigureSubdomain(final SubDomainConfiguration newSubdomainConfiguration)
            throws NoSubdomainConfigurationProvidedException, NoSubdomainNameProvidedException,
            SubdomainAlreadyExistException {
        if (newSubdomainConfiguration == null) {
            throw new NoSubdomainConfigurationProvidedException();
        } else {
            final String newSubdomainName = newSubdomainConfiguration.getName();
            if (newSubdomainName == null) {
                throw new NoSubdomainNameProvidedException();
            } else if (this.subDomainsConfig.get(newSubdomainName) == null) {
                this.subDomainsConfig.remove(this.subDomainConfiguration.getName());
                this.subDomainsConfig.put(newSubdomainName, this.subDomainConfiguration);

                this.subDomainConfiguration.setName(newSubdomainName);
                this.subDomainConfiguration.setDescription(newSubdomainConfiguration.getDescription());
                this.subDomainConfiguration.setMode(newSubdomainConfiguration.getMode());
                this.subDomainConfiguration.setNetworkTimeSynchronized(newSubdomainConfiguration
                        .isNetworkTimeSynchronized());
                this.subDomainConfiguration.setRegistryImplementation(newSubdomainConfiguration
                        .getRegistryImplementation());
                this.subDomainConfiguration.setExtraParameters(newSubdomainConfiguration.getExtraParameters());
            } else {
                throw new SubdomainAlreadyExistException(newSubdomainName);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void reconfigureContainer(final SubDomainConfiguration newSubdomainConfiguration)
            throws NoSubdomainConfigurationProvidedException, NoSubdomainNameProvidedException {
        if (newSubdomainConfiguration == null) {
            throw new NoSubdomainConfigurationProvidedException();
        } else {
            final String newSubdomainName = newSubdomainConfiguration.getName();
            if (newSubdomainName == null) {
                throw new NoSubdomainNameProvidedException();
            } else {
                this.containerConfiguration.setSubdomainName(newSubdomainName);
                this.containersConfig.get(this.containerConfiguration.getName()).setSubdomainName(newSubdomainName);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isSecurityTopologyPassphraseOk(final String securityPassPhrase) {

        // Check security
        final boolean isSecurityOk;
        if (securityPassPhrase == null || securityPassPhrase.isEmpty()) {
            isSecurityOk = false;
        } else {
            final String containerTopologyPassPhrase = this.containerConfiguration.getTopologyPassPhrase();
            if (containerTopologyPassPhrase == null || containerTopologyPassPhrase.isEmpty()
                    || !containerTopologyPassPhrase.equals(securityPassPhrase)) {
                isSecurityOk = false;
            } else {
                isSecurityOk = true;
            }
        }
        return isSecurityOk;
    }

    /**
     * {@inheritDoc}
     */
    public void persistTopology(final Topology topology) throws ConfigurationException {
        final String workingAreaValue = this.containerConfiguration.getWorkDirectoryPath();
        final File workingArea = new File(workingAreaValue);
        if (workingArea.exists()) {
            if (workingArea.isDirectory()) {
                final File persistedTopologyFile = new File(workingArea, PERSISTED_TOPOLOGY_DEFINITION);
                if (persistedTopologyFile.exists()) {
                    if (!persistedTopologyFile.isFile()) {
                        throw new ConfigurationException(String.format(
                                "Invalid persisted topology definition '%s' (not a file).",
                                persistedTopologyFile.getAbsolutePath()));
                    }
                }

                try {
                    TopologyBuilder.marshallTopologyFile(persistedTopologyFile, topology);
                } catch (final TopologyException e) {
                    throw new ConfigurationException(String.format(
                            "An error occurs persisting the topology definition in file '%s'.",
                            persistedTopologyFile.getAbsolutePath()), e);
                }
            } else {
                throw new ConfigurationException(String.format("Invalid working area '%s' (not a directory).",
                        workingAreaValue));
            }
        } else {
            throw new ConfigurationException(String.format("The working area '%s' does not exists.", workingAreaValue));
        }
    }
}
